#https://discourse.panda3d.org/t/using-ssbo-for-light-data/30433/14
#https://discourse.panda3d.org/t/glsl-dithering-filter-help/15010/15
#https://github.com/panda3d/panda3d/issues/630
#https://github.com/ComradeNapCali/libpandabsp/blob/feee2e7d3e102aebf0ed70317d4702f39e324b62/src/libpandabsp/shader_generator.cpp#L153
#https://discourse.panda3d.org/t/a-cheat-sheet-on-matrices-in-the-shader/30054

from panda3d.core import *
load_prc_file_data("", "gl-version 3 1")
load_prc_file_data("", "basic-shaders-only #f")
load_prc_file_data("", """
gl-check-errors True      
gl-debug True    
gl-debug-object-labels True
gl-force-fbo-color False
init-ms-meter
frame-rate-meter-milliseconds true
frame-rate-meter-ms-text-pattern %0.2f ms

#pstats-max-rate 200
pstats-gpu-timing True
pstats-name DemoMaker stats
pstats-python-profiler True
pstats-tasks 1

sync-video False

""")
load_prc_file_data('','textures-power-2 None')
from direct.showbase.ShowBase import ShowBase
from direct.interval.LerpInterval import LerpPosInterval,LerpHprInterval
from direct.interval.IntervalGlobal import *

from ProceduralCube import ProceduralCube,ProceduralCubeTangents

class CubeFaceDef(object):
    """Internal class used by DMKShadowMapGenerator to describe the 6 directions of the camera"""

    def __init__(self,name,look_at,up,face):
        self._name = name
        self._look_at = look_at
        self._up = up
        self._face = face

class Main(ShowBase):

    def __init__(self):

        ShowBase.__init__(self)
        #base.set_background_color(0, 0, 0)
        print(100*"*")
        print("Panda version: ",PandaSystem.getVersionString())
        print(100*"*")
        #Debug
        self.set_frame_rate_meter(True)
        #PStatClient.connect()
        #End of debug
        light = DirectionalLight("MyDirectional")
        light.color = (0.0,0.7,0.3,1.0)
        self._light_np = base.render.attach_new_node(light)
        self._light_np.set_pos(0,-500,10.0)
        self._light_np.node().set_shadow_caster(True,512,512,-2)
        self._light_np.node().get_lens().set_film_size(500.0,500.0)
        self._light_np.node().get_lens().set_near_far(1.0,1000.0)
        base.render.set_light(self._light_np)

        light2 = Spotlight("MySpot")
        light2.color = (5.0,0.1,0.0,1.0)
        light2.max_distance = 8000.0
        self._light2_np = base.render.attach_new_node(light2)
        self._light2_np.set_pos(0,0,25)
        self._light2_np.node().set_shadow_caster(True,512,512,-1)
        base.render.set_light(self._light2_np)

        light3 = PointLight("MyPL")
        light3.color = (0.0,0.1,5.0,1.0)
        light3.max_distance = 5000.0
        light3.priority = 10
        light3.setColorTemperature( 3600 )
        self._light3_np = base.render.attach_new_node(light3)
        self._light3_np.node().set_shadow_caster(True,512,512,-2)
        self._light3_np.set_pos(200,-100,-80.0)
        for index in range(6):
            self._light3_np.node().get_lens(index).set_near_far(1.0,500.0)
        base.render.set_light(self._light3_np)
        #light3.show_frustum()
        self._light3_np.set_light_off()

        light4 = PointLight("MyPL2")
        light4.color = (5.0,0.1,0.0,1.0)
        light4.max_distance = 5000.0
        light4.priority = 15
        light4.setColorTemperature( 3600 )
        self._light4_np = base.render.attach_new_node(light4)
        self._light4_np.node().set_shadow_caster(True,512,512,-1)
        self._light4_np.set_pos(200,-100,-180.0)
        for index in range(6):
            self._light4_np.node().get_lens(index).set_near_far(1.0,500.0)
        base.render.set_light(self._light4_np)
        self._light4_np.set_light_off()

        self._np = base.loader.load_model(Filename.from_os_specific("box"))
        self._np.reparent_to(base.render)
        self._np.set_two_sided(False)

        max_nb_instances = 10
        instanced_np = NodePath("Instanced_NP")
        for x in range (0,max_nb_instances):
            for y in range(0,max_nb_instances):
                for z in range(0,max_nb_instances):
                    np2 = base.loader.load_model(Filename.from_os_specific("box"))
                    np2.reparent_to(instanced_np)
                    np2.set_two_sided(False)
                    np2.set_pos(2*x*15,-10.0*y*2,-30.0*z*2)
                    np2.set_scale(15.0)
        instanced_np.reparent_to(base.render)
        #Comment it out to speed up
        #instanced_np.clear_model_nodes()
        #instanced_np.flatten_strong()

        pl_impostor = base.loader.load_model(Filename.from_os_specific("jack"))
        pl_impostor.reparent_to(self._light3_np)
        pl_impostor.set_scale(5.0)
        pl_impostor.set_pos(0.0,50.0,0.0)

        pl_impostor2 = base.loader.load_model(Filename.from_os_specific("jack"))
        pl_impostor2.reparent_to(self._light4_np)
        pl_impostor2.set_pos(0.0,0.0,20.0)
        pl_impostor2.set_scale(1.0)

        pc = ProceduralCube()
        snode = pc.GenerateNode()
        self._pl_cube_renderer = base.render.attach_new_node(snode)
        self._pl_cube_renderer.set_two_sided(True)
        self._pl_cube_renderer.set_scale(80.0)
        self._pl_cube_renderer.set_light_off()
        self._pl_cube_renderer.set_fog_off()
        self._pl_cube_renderer.set_color_off()
        self._pl_cube_renderer.reparent_to(base.render)

        pc = ProceduralCube()
        snode = pc.GenerateNode()
        self._pl_cube_renderer2 = base.render.attach_new_node(snode)
        self._pl_cube_renderer2.set_pos(150.0,0.0,0.0)
        self._pl_cube_renderer2.set_two_sided(True)
        self._pl_cube_renderer2.set_scale(45.0)
        self._pl_cube_renderer2.set_light_off()
        self._pl_cube_renderer2.set_fog_off()
        self._pl_cube_renderer2.set_color_off()
        self._pl_cube_renderer2.reparent_to(base.render)

        self._np.set_shader_auto(True)

        self._light2_np.look_at(self._np)
        self._light_np.look_at(self._np)

        self._tex_array = Texture("volume")
        self._tex_array.setup_2d_texture_array(512, 512, 4, Texture.T_float, Texture.F_depth_component32)
        self._tex_array.set_clear_color( LVecBase4(0.0,0.0,0.0,1.0 ))
        self._state = SamplerState()
        self._state.setMinfilter( SamplerState.FT_linear)
        self._state.setMagfilter( SamplerState.FT_linear )
        self._state.set_wrap_u( SamplerState.WM_clamp )
        self._state.set_wrap_v( SamplerState.WM_clamp )
        self._state.set_wrap_w( SamplerState.WM_clamp )
        self._tex_array.setDefaultSampler( self._state )

        self._tex_array2 = Texture("volume2")
        #Works for 2 textures (instead of 12) - 1 makes the same texture to be displayed so 2 is minimum
        self._tex_array2.setup_cube_map_array(512,2, Texture.T_float, Texture.F_depth_component32)
        self._tex_array2.set_clear_color( LVecBase4( 1.0 ))
        self._state = SamplerState()
        #FT_linear_mipmap_linear does not work
        self._state.set_minfilter(SamplerState.FT_linear)
        self._state.set_magfilter(SamplerState.FT_linear)
        self._state.set_wrap_u( SamplerState.WM_clamp )
        self._state.set_wrap_v( SamplerState.WM_clamp )
        self._state.set_wrap_w( SamplerState.WM_clamp )
        self._tex_array2.set_default_sampler( self._state )

        self._start_pos = self._light2_np.get_pos()
        self._end_pos = Point3(Vec3(self._start_pos.get_x()-10.0,self._start_pos.get_y(),self._start_pos.get_z()+50.0) ) 
        interval1 = LerpPosInterval(self._light2_np,5.0,pos = self._end_pos,startPos = self._start_pos) 
        interval2 = LerpPosInterval(self._light2_np,5.0,pos = self._start_pos)
        self._seq = Sequence(interval1, interval2,name="MySequence")
        self._seq.loop()

        self._start_pos = self._light3_np.get_pos()
        self._end_pos = Point3(Vec3(self._start_pos.get_x()+100,self._start_pos.get_y(),self._start_pos.get_z()-200.0) ) 
        interval3 = LerpPosInterval(self._light3_np,5.0,pos = self._end_pos,startPos = self._start_pos)
        interval3.loop()
 
        interval4 = LerpHprInterval(self._light4_np,5.0,hpr = (360.0,360.0,360.0),startHpr = (0.0,0.0,0.0))
        interval4.loop()
 
        shader = self._generate_cube_3D_shader()
        self._pl_cube_renderer.set_shader(shader)
        self._pl_cube_renderer.set_shader_input("texture_array3d",self._tex_array2)
        self._pl_cube_renderer.set_shader_input("DMKlayer",0)
        self._pl_cube_renderer2.set_shader(shader)
        self._pl_cube_renderer2.set_shader_input("texture_array3d",self._tex_array2)
        self._pl_cube_renderer2.set_shader_input("DMKlayer",1)

        self.accept("v", base.bufferViewer.toggleEnable)
        base.taskMgr.add(self._capture)


    def _capture(self,task):
        sBuffer = self._light_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))
        sBuffer2 = self._light2_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))
        sBuffer3 = self._light3_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))
        sBuffer4 = self._light4_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))

        if sBuffer is not None and sBuffer2 is not None and sBuffer3 is not None and sBuffer4 is not None:
            self.cardmaker = CardMaker("cubemaker")
            self.cardmaker.setFrame(-1,1,-1,1)
            for layer in range(3):
                if layer in [0,1]: 
                    self.cardmaker.setUvRange((0, 1, 1, 0), (0, 0, 1, 1),(layer, layer, layer, layer))
                    self._create_card_2D(layer,0)
            #Light1
            sBuffer.clear_render_textures()
            sBuffer.add_render_texture(self._tex_array, GraphicsOutput.RTM_bind_or_copy, GraphicsOutput.RTP_depth)
            for dr in sBuffer.get_display_regions():
                dr.set_target_tex_page(0)
            #Light2
            dr = sBuffer.make_mono_display_region((0, 1, 0, 1))
            dr.set_target_tex_page(1)
            dr.set_camera(self._light2_np)
            dr.set_clear_depth_active(True)
            dr.set_active(True)
            sBuffer2.set_active(False)
            #Light3
            sBuffer3.clear_render_textures()
            sBuffer3.add_render_texture(self._tex_array2, GraphicsOutput.RTM_bind_or_copy, GraphicsOutput.RTP_depth)
            layer = 0
            face = 0
            for dr in sBuffer3.get_display_regions():
                if dr.active is True:
                    dr.set_target_tex_page(6*layer + face)
                    face +=1
            #Light4
            layer = 1
            face = 0
            cube_faces = [CubeFaceDef("positive_x", LPoint3(1, 0, 0), LVector3(0, -1, 0),0),
        CubeFaceDef("negative_x", LPoint3(-1, 0, 0), LVector3(0, -1, 0),1),
        CubeFaceDef("positive_y", LPoint3(0, 1, 0), LVector3(0, 0, 1),2),
        CubeFaceDef("negative_y", LPoint3(0, -1, 0), LVector3(0, 0, -1),3),
        CubeFaceDef("positive_z", LPoint3(0, 0, 1), LVector3(0, -1, 0),4),
        CubeFaceDef("negative_z", LPoint3(0, 0, -1), LVector3(0, -1, 0),5)]
            for face in range(0,6):
                dr = sBuffer3.make_mono_display_region((0, 1, 0, 1))
                dr.set_target_tex_page(6*layer + face)
                camera = Camera(cube_faces[face]._name)
                camera_np = self._light4_np.attach_new_node(camera)
                camera_np.look_at(cube_faces[face]._look_at, cube_faces[face]._up)
                dr.set_camera(camera_np)
                dr.set_clear_depth_active(True)
                dr.set_active(True)
            sBuffer4.set_active(False)
            return task.done
        return task.again

    def _create_card_2D(self,layer,shift):
        """Helper method to create the 2D debug card"""
        card = NodePath(self.cardmaker.generate())
        card.set_two_sided(True)
        card.set_scale(0.1)           
        card.set_pos(-0.9 + 0.25*(layer+shift),0.0,0.6)
        shader = self._generate_card_2D_shader()
        card.set_shader(shader)
        card.set_shader_input("texture_array",self._tex_array)
        card.set_shader_input("DMKlayer",layer)
        card.reparent_to(base.render2d)

    def _generate_card_2D_shader(self): 
        """Debug shader for 2D texture"""   
        file = open("DebugShaders/vertex_VS.glsl")
        vs = file.read()
        file = open("DebugShaders/fragment_FS.glsl")
        fs = file.read()
        return Shader.make(Shader.SL_GLSL, vertex = vs, fragment = fs)

    def _generate_cube_3D_shader(self):
        """Debug shader for 3d texture"""
        file = open("DebugShaders/vertex_3dTexture_VS.glsl")
        vs = file.read()
        file = open("DebugShaders/fragment_3dTexture_FS.glsl")
        fs = file.read()
        return Shader.make(Shader.SL_GLSL, vertex = vs, fragment = fs)

    def _generate_shadow_shader(self):
        file = open("ShadowShaders/vertex_shadow_VS.glsl")
        vs = file.read()
        file = open("ShadowShaders/fragment_shadow_FS.glsl")
        fs = file.read()
        shader =  Shader.make(Shader.SL_GLSL, vertex = vs, fragment = fs)
        shader.set_filename(0,"Light Shader")
        attr = ShaderAttrib.make(shader)
        return attr

a = Main()
a.run()

